/*
 * Copyright (c) 1991-2023 by STEP Tools Inc. 
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and
 * its documentation is hereby granted, provided that this copyright
 * notice and license appear on all copies of the software.
 * 
 * STEP TOOLS MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE
 * SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING
 * BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT. STEP TOOLS
 * SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A
 * RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES.
 *
 * Author: Dave Loffredo (loffredo@steptools.com)
 */

#include <structural_frame_schema.h>
#include "unit_utils.h"

// Construct the various common SI and English units.  Note that CIS/2
// usage handles english/degree units differently than STEP.  Instead
// of using conversion-based unit, CIS/2 just uses a context-dependent
// unit with the appropriate name.
//


/* PRINT UNIT NAME -- STEP units are usually represented as a complex
 * instance composed of named unit types.  The usual set of units are
 * the SI units (meter, gram, second), although it is posible to define
 * conversion-based units such as inches, pounds, etc.  It is also
 * possible to define arbitrary contxt-dependant units such as scoops,
 * rolls, etc.
 */
void print_unit (unit * u)
{
    if (!u) return;
    if (u-> is_named_unit()) {
	named_unit * nu = u-> _named_unit();
	if (nu-> isa (ROSE_TYPE(si_unit))) {
	    si_unit * su = ROSE_CAST (si_unit, nu);
	    switch (su->prefix()) {
	    case si_prefix_EXA:		printf ("exa");		break;
	    case si_prefix_PICO:	printf ("pico");	break;
	    case si_prefix_MEGA:	printf ("mega");	break;
	    case si_prefix_FEMTO:	printf ("femto");	break;
	    case si_prefix_ATTO:	printf ("atto");	break;
	    case si_prefix_CENTI:	printf ("centi");	break;
	    case si_prefix_NANO:	printf ("nano");	break;
	    case si_prefix_HECTO:	printf ("hecto");	break;
	    case si_prefix_MICRO:	printf ("micro");	break;
	    case si_prefix_TERA:	printf ("tera");	break;
	    case si_prefix_GIGA:	printf ("giga");	break;
	    case si_prefix_MILLI:	printf ("milli");	break;
	    case si_prefix_PETA:	printf ("peta");	break;
	    case si_prefix_DECI:	printf ("deci");	break;
	    case si_prefix_KILO:	printf ("kilo");	break;
	    case si_prefix_DECA:	printf ("deca");	break;
	    case si_prefix_NULL:	
	    default:
		/* nothing */ break;
	    }

	    switch (su->name()) {
	    case si_unit_name_HERTZ:	printf ("hertz");	break;
	    case si_unit_name_DEGREE_CELSIUS: printf ("degree_celsius"); break;
	    case si_unit_name_SIEMENS:	printf ("siemens");	break;
	    case si_unit_name_SIEVERT:	printf ("sievert");	break;
	    case si_unit_name_LUX:	printf ("lux");		break;
	    case si_unit_name_WATT:	printf ("watt");	break;
	    case si_unit_name_OHM:	printf ("ohm");		break;
	    case si_unit_name_SECOND:	printf ("second");	break;
	    case si_unit_name_BECQUEREL:printf ("becquerel");	break;
	    case si_unit_name_PASCAL:	printf ("pascal");	break;
	    case si_unit_name_HENRY:	printf ("henry");	break;
	    case si_unit_name_TESLA:	printf ("tesla");	break;
	    case si_unit_name_VOLT:	printf ("volt");	break;
	    case si_unit_name_JOULE:	printf ("joule");	break;
	    case si_unit_name_KELVIN:	printf ("kelvin");	break;
	    case si_unit_name_AMPERE:	printf ("ampere");	break;
	    case si_unit_name_GRAM:	printf ("gram");	break;
	    case si_unit_name_STERADIAN:printf ("steradian");	break;
	    case si_unit_name_MOLE:	printf ("mole");	break;
	    case si_unit_name_LUMEN:	printf ("lumen");	break;
	    case si_unit_name_GRAY:	printf ("gray");	break;
	    case si_unit_name_CANDELA:	printf ("candela");	break;
	    case si_unit_name_FARAD:	printf ("farad");	break;
	    case si_unit_name_RADIAN:	printf ("radian");	break;
	    case si_unit_name_NEWTON:	printf ("newton");	break;
	    case si_unit_name_METRE:	printf ("metre");	break;
	    case si_unit_name_WEBER:	printf ("weber");	break;
	    case si_unit_name_COULOMB:	printf ("coulomb");	break;
	    case si_unit_name_NULL:	
	    default:
		printf ("none");
		break;
	    }
	}
	if (nu-> isa (ROSE_TYPE(conversion_based_unit))) {
	    conversion_based_unit * cbu = ROSE_CAST(conversion_based_unit,nu);
	    printf ("%s (conversion-based unit)", 
		    (cbu->name()? cbu->name(): "none"));
	}
	if (nu-> isa (ROSE_TYPE(context_dependent_unit))) {
	    context_dependent_unit * cdu = ROSE_CAST(context_dependent_unit,nu);
	    printf ("%s (context-dependent unit)", 
		    (cdu->name()? cdu->name(): "none"));
	}
    }
}



//------------------------------------------------------------
// Construct the various common SI and English units.
//
// These functions only create one instance of a given unit in a
// design.  The first time the unit is created, we save a shortcut in
// the nametable of the design.  This lets us quickly find the unit
// again if we need it, which keeps us from creating duplicate unit
// definitions.
//
named_unit * make_meter_unit(
    RoseDesign * d
    )
{
    const char * shortcut = "__CONSTANT UNIT metre";

    if (!d) d = ROSE.design();

    /* Use shortcut to unit if we have already created one */
    RoseObject * obj = d-> findObject (shortcut);
    if (obj) return ROSE_CAST (named_unit, obj);

    // Need a complex instance of both si and length unit
    si_unit * u = pnewIn(d) length_unit_and_si_unit;
    u-> name (si_unit_name_METRE);
    u-> prefix (si_prefix_NULL);

    d-> addName (shortcut, u);  // save shortcut for future calls
    return u;
}

named_unit * make_mm_unit(
    RoseDesign * d
    )
{
    const char * shortcut = "__CONSTANT UNIT millimetre";

    if (!d) d = ROSE.design();

    /* Use shortcut to unit if we have already created one */
    RoseObject * obj = d-> findObject (shortcut);
    if (obj) return ROSE_CAST (named_unit, obj);

    // Need a complex instance of both si and length unit
    si_unit * u = pnewIn(d) length_unit_and_si_unit;
    u-> name (si_unit_name_METRE);
    u-> prefix (si_prefix_MILLI);

    d-> addName (shortcut, u);  // save shortcut for future calls
    return u;
}

named_unit * make_foot_unit(
    RoseDesign * d
    )
{
    const char * shortcut = "__CONSTANT UNIT foot";

    if (!d) d = ROSE.design();

    /* Use shortcut to unit if we have already created one */
    RoseObject * obj = d-> findObject (shortcut);
    if (obj) return ROSE_CAST (named_unit, obj);

    // CIS/2 usage seems to be to use context-dependent units
    context_dependent_unit * u = 
	pnewIn(d) context_dependent_unit_and_length_unit;

    u-> name ("FOOT");
    u-> dimensions (pnewIn(d) dimensional_exponents (1,0,0,0,0,0,0));
    d-> addName (shortcut, u);  // save shortcut for future calls
    return u;
}
// Make a conversion based-inch unit as is done in STEP
named_unit * make_foot_unit_as_cbu(
    RoseDesign * d
    )
{
    const char * shortcut = "__CONSTANT UNIT inch";

    if (!d) d = ROSE.design();

    /* Use shortcut to unit if we have already created one */
    RoseObject * obj = d-> findObject (shortcut);
    if (obj) return ROSE_CAST (named_unit, obj);

    // Inch is defined as a conversion-based unit in terms of an
    // underlying SI unit.  Usually 25.4mm or 2.54cm.  Create the
    // measure with unit for foot as 12x this.  
    //
    measure_with_unit * mwu = pnewIn(d) measure_with_unit;
    mwu-> value_component (pnewIn(d) measure_value);
    mwu-> value_component ()-> _length_measure (25.4 * 12);

    mwu-> unit_component (pnewIn(d) unit);
    mwu-> unit_component ()-> _named_unit (make_mm_unit(d));


    // The unit itself needs a complex instance of both CBU and length
    // unit.  Conversion based unit also needs dimensional exponents
    // that show which basic quantities (length, time, mass, etc) the
    // unit describes.
    //
    conversion_based_unit * u = 
	pnewIn(d) conversion_based_unit_and_length_unit;

    u-> name ("FOOT");
    u-> dimensions (pnewIn(d) dimensional_exponents (1,0,0,0,0,0,0));
    u-> conversion_factor (mwu);

    d-> addName (shortcut, u);  // save shortcut for future calls
    return u;
}




named_unit * make_inch_unit(
    RoseDesign * d
    )
{
    const char * shortcut = "__CONSTANT UNIT inch";

    if (!d) d = ROSE.design();

    /* Use shortcut to unit if we have already created one */
    RoseObject * obj = d-> findObject (shortcut);
    if (obj) return ROSE_CAST (named_unit, obj);

    // CIS/2 usage seems to be to use context-dependent units
    context_dependent_unit * u = 
	pnewIn(d) context_dependent_unit_and_length_unit;

    u-> name ("INCH");
    u-> dimensions (pnewIn(d) dimensional_exponents (1,0,0,0,0,0,0));
    d-> addName (shortcut, u);  // save shortcut for future calls
    return u;
}



// Make a conversion based-inch unit as is done in STEP
named_unit * make_inch_unit_as_cbu(
    RoseDesign * d
    )
{
    const char * shortcut = "__CONSTANT UNIT inch";

    if (!d) d = ROSE.design();

    /* Use shortcut to unit if we have already created one */
    RoseObject * obj = d-> findObject (shortcut);
    if (obj) return ROSE_CAST (named_unit, obj);

    // Inch is defined as a conversion-based unit in terms of an
    // underlying SI unit.  Usually 25.4mm or 2.54cm.  Create the
    // measure with unit to indicate this.  
    //
    measure_with_unit * mwu = pnewIn(d) measure_with_unit;
    mwu-> value_component (pnewIn(d) measure_value);
    mwu-> value_component ()-> _length_measure (25.4);

    mwu-> unit_component (pnewIn(d) unit);
    mwu-> unit_component ()-> _named_unit (make_mm_unit(d));


    // The unit itself needs a complex instance of both CBU and length
    // unit.  Conversion based unit also needs dimensional exponents
    // that show which basic quantities (length, time, mass, etc) the
    // unit describes.
    //
    conversion_based_unit * u = 
	pnewIn(d) conversion_based_unit_and_length_unit;

    u-> name ("INCH");
    u-> dimensions (pnewIn(d) dimensional_exponents (1,0,0,0,0,0,0));
    u-> conversion_factor (mwu);

    d-> addName (shortcut, u);  // save shortcut for future calls
    return u;
}



named_unit * make_radian_unit(
    RoseDesign * d
    )
{
    const char * shortcut = "__CONSTANT UNIT radian";

    if (!d) d = ROSE.design();

    /* Use shortcut to unit if we have already created one */
    RoseObject * obj = d-> findObject (shortcut);
    if (obj) return ROSE_CAST (named_unit, obj);

    // Need a complex instance of both si and plane angle unit
    si_unit * u = pnewIn(d) plane_angle_unit_and_si_unit;
    u-> name (si_unit_name_RADIAN);
    u-> prefix (si_prefix_NULL);

    d-> addName (shortcut, u);  // save shortcut for future calls
    return u;
}



named_unit * make_degree_unit(
    RoseDesign * d
    )
{
    const char * shortcut = "__CONSTANT UNIT degree";

    if (!d) d = ROSE.design();

    /* Use shortcut to unit if we have already created one */
    RoseObject * obj = d-> findObject (shortcut);
    if (obj) return ROSE_CAST (named_unit, obj);

    // CIS/2 usage seems to be to use context-dependent units
    context_dependent_unit * u = 
	pnewIn(d) context_dependent_unit_and_plane_angle_unit;

    u-> name ("DEGREE");
    u-> dimensions (pnewIn(d) dimensional_exponents (0,0,0,0,0,0,0));
    d-> addName (shortcut, u);  // save shortcut for future calls
    return u;
}


// Make a conversion-based degree unit as is done in STEP
named_unit * make_degree_unit_as_cbu(
    RoseDesign * d
    )
{
    const char * shortcut = "__CONSTANT UNIT degree";

    if (!d) d = ROSE.design();

    /* Use shortcut to unit if we have already created one */
    RoseObject * obj = d-> findObject (shortcut);
    if (obj) return ROSE_CAST (named_unit, obj);

    // Degree is defined as a conversion-based unit in terms of the
    // underlying radian SI unit.  1 Degree == PI/180 radians.  Create
    // the measure with unit to indicate this.
    //
    measure_with_unit * mwu = pnewIn(d) measure_with_unit;
    mwu-> value_component (pnewIn(d) measure_value);
    mwu-> value_component ()-> _plane_angle_measure (0.01745329252);

    mwu-> unit_component (pnewIn(d) unit);
    mwu-> unit_component ()-> _named_unit (make_radian_unit(d));


    // The unit itself needs a complex instance of both CBU and plane
    // angle unit.  Conversion based unit also needs dimensional
    // exponents but these are all zero for angles.
    //
    conversion_based_unit * u = 
	pnewIn(d) conversion_based_unit_and_plane_angle_unit;

    u-> name ("DEGREE");
    u-> dimensions (pnewIn(d) dimensional_exponents (0,0,0,0,0,0,0));
    u-> conversion_factor (mwu);

    d-> addName (shortcut, u);  // save shortcut for future calls
    return u;
}



named_unit * make_steradian_unit(
    RoseDesign * d
    )
{
    const char * shortcut = "__CONSTANT UNIT steradian";

    if (!d) d = ROSE.design();

    /* Use shortcut to unit if we have already created one */
    RoseObject * obj = d-> findObject (shortcut);
    if (obj) return ROSE_CAST (named_unit, obj);

    // Need a complex instance of both si and length unit
    si_unit * u = pnewIn(d) si_unit_and_solid_angle_unit;
    u-> name (si_unit_name_STERADIAN);
    u-> prefix (si_prefix_NULL);

    d-> addName (shortcut, u);  // save shortcut for future calls
    return u;
}



